/* ADC driver for A20 SoC
 *
 * Copyright (c) 2013 Grandstream Networks, Inc.
 * By Fanan Chen <fnchen@grandstream.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option)any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/sched.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <mach/adc.h>

#define TP_CTRL0                        (0x00)
#define TP_CTRL1                        (0x04)
#define TP_CTRL2                        (0x08)
#define TP_CTRL3                        (0x0c)
#define TP_INT_FIFOC                    (0x10)
#define TP_INT_FIFOS                    (0x14)
#define TP_TPR                          (0x18)
#define TP_CDAT                         (0x1c)
#define TEMP_DATA                       (0x20)
#define TP_DATA                         (0x24)

#define TP_FIFO_FLUSH					(1<<4)
#define FIFO_DATA_PENDING               (1<<16)

#define MAX_CHANNEL	4
#define ADC_TIMER_DELAY	10

struct sunxi_adc_device {
	int irq;
	void __iomem *base;
	unsigned int data[MAX_CHANNEL];
	struct work_struct work;
};

static LIST_HEAD(adc_clients);
static DEFINE_MUTEX(adc_lock);

struct sunxi_adc_client *sunxi_adc_client_register(struct platform_device *pdev,
		void (*convert_cb)(struct sunxi_adc_client *client),
		unsigned char channel, unsigned int debouce)
{
	struct sunxi_adc_client *client, *p;

	if (!pdev)
		return NULL;

	mutex_lock(&adc_lock);
	list_for_each_entry(p, &adc_clients, list) {
		if (p->channel == channel) {
			dev_err(&pdev->dev, "ADC channel%d already registed\n", channel);
			return NULL;
		}
	}

	client = kzalloc(sizeof(struct sunxi_adc_client), GFP_KERNEL);
	if (!client) {
		dev_err(&pdev->dev, "ADC client alloc memory failed\n");
		return NULL;
	}

	client->pdev = pdev;
	client->convert_cb = convert_cb;
	client->channel = channel;
	client->debouce = debouce;
	client->dcount = 0;

	list_add_tail(&client->list, &adc_clients);
	mutex_unlock(&adc_lock);

	return client;
}
EXPORT_SYMBOL_GPL(sunxi_adc_client_register);

void sunxi_adc_client_unregister(struct sunxi_adc_client *client)
{
	struct sunxi_adc_client *tmp = NULL;

	if (!client)
		return;

	mutex_lock(&adc_lock);
	list_for_each_entry(tmp, &adc_clients, list) {
		if (tmp == client)
			list_del(&tmp->list);
	}
	mutex_unlock(&adc_lock);

	kfree(client);
}
EXPORT_SYMBOL_GPL(sunxi_adc_client_unregister);

static void sunxi_adc_work(struct work_struct *work)
{
	struct sunxi_adc_device *adc =
			container_of(work, struct sunxi_adc_device, work);
	struct sunxi_adc_client *p;

	list_for_each_entry(p, &adc_clients, list) {
		if (p->debouce && (p->dcount < p->debouce)) {
			p->dcount++;
			continue;
		}
		if (p->convert_cb) {
			p->dcount = 0;
			p->conv_val = adc->data[p->channel];
			p->convert_cb(p);
		}
	}
}

static void adc_device_init(struct sunxi_adc_device *adc)
{
	unsigned int reg_val;

	writel(0x0080002f, adc->base + TP_CTRL0);
	reg_val = readl(adc->base + TP_CTRL1);
	reg_val |= (1 << 3) | (1 << 4) | 0x4;
	writel(reg_val, adc->base + TP_CTRL1);

	reg_val = (1 << 16) | (0x3 << 8) | (1 << 4);
	writel(reg_val, adc->base + TP_INT_FIFOC);
}

static irqreturn_t adc_irq_handler(int irq, void *dev_id)
{
	struct sunxi_adc_device *adc = (struct sunxi_adc_device *)dev_id;
	unsigned int reg_val, reg_fifoc;
	int i = 0;

	reg_val = readl(adc->base + TP_INT_FIFOS);
	if (reg_val & FIFO_DATA_PENDING) {
		for (i = 0; i < MAX_CHANNEL; i++)
			adc->data[i] = readl(adc->base + TP_DATA);

		/* flush fifo */
		reg_fifoc = readl(adc->base + TP_INT_FIFOC);
		reg_fifoc |= TP_FIFO_FLUSH;
		writel(reg_fifoc, adc->base + TP_INT_FIFOC);
		/* clear pending */
		writel(reg_val | FIFO_DATA_PENDING, adc->base + TP_INT_FIFOS);

		schedule_work(&adc->work);
	}

	return IRQ_HANDLED;
}

static int __devinit sunxi_adc_probe(struct platform_device *pdev)
{
	int ret = 0;
	struct resource *res = NULL;
	struct sunxi_adc_device *adc = NULL;

	adc = kzalloc(sizeof(struct sunxi_adc_device), GFP_KERNEL);
	if (!adc) {
		dev_err(&pdev->dev, "Alloc memory for adc failed\n");
		return -ENOMEM;
	}

	adc->irq = platform_get_irq(pdev, 0);
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	adc->base = ioremap(res->start, resource_size(res));
	if (!adc->base) {
		dev_err(&pdev->dev, "Failed to map registers.\n");
		ret = -ENXIO;
		goto err_ioremap;
	}

	adc_device_init(adc);

	INIT_WORK(&adc->work, sunxi_adc_work);

	platform_set_drvdata(pdev, adc);

	ret = request_irq(adc->irq, adc_irq_handler,
			IRQF_DISABLED, pdev->name, adc);

	return 0;

err_irq:
	iounmap(adc->base);
err_ioremap:
	kfree(adc);
	return ret;
}

static int __devexit sunxi_adc_remove(struct platform_device *pdev)
{
	struct sunxi_adc_device *adc = platform_get_drvdata(pdev);

	free_irq(adc->irq, adc);
	iounmap(adc->base);
	kfree(adc);

	return 0;
}

static struct platform_driver sunxi_adc_driver = {
	.driver		= {
		.name	= "sunxi-adc",
		.owner	= THIS_MODULE,
	},
	.probe		= sunxi_adc_probe,
	.remove		= __devexit_p(sunxi_adc_remove),
};

static int __init sunxi_adc_init(void)
{
	return platform_driver_register(&sunxi_adc_driver);
}

module_init(sunxi_adc_init);
