/*
 * PWM driver for A20 SoC chip.
 *
 * Copyright (c) 2013 Grandstream Networks, Inc.
 * By Fanan Chen <fnchen@grandstream.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option)any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/pwm.h>
#include <linux/export.h>
#include <linux/err.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/io.h>

struct pwm_device {
	struct list_head list;
	struct platform_device *pdev;
	void __iomem *regbase;
	const char *label;
	unsigned int pwm_id;
	unsigned int use_count;
	unsigned int period_ns;
	unsigned int duty_ns;
};

enum sunxi_pwm_prescale {
	PRESCALE_DIV120  = 0x00,
	PRESCALE_DIV180  = 0x01,
	PRESCALE_DIV240  = 0x02,
	PRESCALE_DIV360  = 0x03,
	PRESCALE_DIV480  = 0x04,
	PRESCALE_INVx05  = 0x05,
	PRESCALE_INVx06  = 0x06,
	PRESCALE_INVx07  = 0x07,
	PRESCALE_DIV12k  = 0x08,
	PRESCALE_DIV24k  = 0x09,
	PRESCALE_DIV36k  = 0x0a,
	PRESCALE_DIV48k  = 0x0b,
	PRESCALE_DIV72k  = 0x0c
};

static DEFINE_MUTEX(pwm_lock);
static LIST_HEAD(pwm_list);

#define PWMS_NR	2
#define MAX_CYCLES	0x0ffff
#define MIN_CYCLES	0x02

#define PWM_CTL_REG		0x200
#define PWM_CH0_PERIOD	0x204
#define PWM_CH1_PERIOD	0x208

struct pwm_device *pwm_request(int pwm_id, const char *label)
{
	struct pwm_device *pwm;
	int found = 0;

	mutex_lock(&pwm_lock);

	list_for_each_entry(pwm, &pwm_list, list)
	{
		if (pwm->pwm_id == pwm_id)
		{
			found = 1;
			break;
		}
	}

	if (found)
	{
		if (pwm->use_count == 0)
		{
			pwm->use_count = 1;
			pwm->label = label;
		}
		else
			pwm = ERR_PTR(-EBUSY);
	}
	else
		pwm = ERR_PTR(-ENOENT);

	mutex_unlock(&pwm_lock);

	return pwm;
}
EXPORT_SYMBOL(pwm_request);

void pwm_free(struct pwm_device *pwm)
{
	mutex_lock(&pwm_lock);

	if (pwm->use_count) {
		pwm->use_count--;
		pwm->label = NULL;
	} else
		printk(KERN_ERR "PWM%d device already freed\n", pwm->pwm_id);

	mutex_unlock(&pwm_lock);
}
EXPORT_SYMBOL(pwm_free);

#define msecs_to_loops(t) (loops_per_jiffy / 1000 * HZ * t)
static inline void pwm_busy_wait(struct pwm_device *pwm)
{
	int loops = msecs_to_loops(10);
	__u32 mask = 1 << (28 + pwm->pwm_id);

	while ((readl(pwm->regbase + PWM_CTL_REG)&mask)
			&& --loops)
		cpu_relax();

	if (unlikely(!loops))
		printk(KERN_WARNING "pwm wait time out\n");
}

static const unsigned int prescale_divisor[13] = {120,
						  180,
						  240,
						  360,
						  480,
						  480, /* Invalid Option */
						  480, /* Invalid Option */
						  480, /* Invalid Option */
						  12000,
						  24000,
						  36000,
						  48000,
						  72000};

enum sunxi_pwm_prescale  pwm_get_best_prescale(unsigned int period_in) {
	int i;
	unsigned int period = period_in;
	enum sunxi_pwm_prescale best_prescale = 0;

	best_prescale = -1;
	for(i = 0 ; i < 13 ; i++) {
		unsigned int check_value = (prescale_divisor[i] /24);
		if(check_value < 1 || check_value > period) {
			break;
		}
		if(((period / check_value) >= (MAX_CYCLES / 2)) &&
			((period / check_value) <= MAX_CYCLES)) {
			best_prescale = i;
			break;
		}
	}

	if(best_prescale > 13) {
		for(i = 0 ; i < 13 ; i++) {
			unsigned int check_value = (prescale_divisor[i] /24);
			if(check_value < 1 || check_value > period) {
				break;
			}
			if(((period / check_value) >= MIN_CYCLES) &&
				((period / check_value) <= MAX_CYCLES)) {
				best_prescale = i;
				break;
			}
		}
	}
	if(best_prescale > 13) {
		best_prescale = PRESCALE_DIV480;  /* Something that's not zero - use invalid prescale value */
	}

	return best_prescale;
}

int pwm_config(struct pwm_device *pwm, int duty_ns, int period_ns)
{
	int i = 0;
	__u32 value = 0;
	unsigned int duty = duty_ns / 1000;
	unsigned int period = period_ns / 1000;
    unsigned int prescale = 0, cyc_cnt = 0;
	unsigned int entire_cyc = 0, active_cyc = 0;

	if (pwm == NULL || period_ns == 0 || duty_ns > period_ns)
		return -EINVAL;

	prescale = pwm_get_best_prescale(period);
	entire_cyc = period / (prescale_divisor[prescale]/24);
	if (entire_cyc > MAX_CYCLES)
		entire_cyc = MAX_CYCLES;
	active_cyc = duty / (prescale_divisor[prescale]/24);
	if (entire_cyc > MAX_CYCLES)
		active_cyc = entire_cyc - 1;

	value = readl(pwm->regbase + PWM_CTL_REG);
	switch (pwm->pwm_id)
	{
		case 0:
			value |= (prescale << 0);
			writel(value, pwm->regbase + PWM_CTL_REG);
			value = (active_cyc&0xffff)|((entire_cyc&0xffff) << 16);
			pwm_busy_wait(pwm);
			writel(value, pwm->regbase + PWM_CH0_PERIOD);
			break;
		case 1:
			value |= (prescale << 15)|(3 << 20);
			writel(value, pwm->regbase + PWM_CTL_REG);
			value = (active_cyc&0xffff)|((entire_cyc&0xffff) << 16);
			pwm_busy_wait(pwm);
			writel(value, pwm->regbase + PWM_CH1_PERIOD);
			break;
		default:
			return -EINVAL;
	}

	pwm->period_ns = period_ns;
	pwm->duty_ns = duty_ns;

	return 0;
}
EXPORT_SYMBOL(pwm_config);

int pwm_enable(struct pwm_device *pwm)
{
	u32 value = readl(pwm->regbase + PWM_CTL_REG);
	switch (pwm->pwm_id)
	{
		case 0:
			value |= 0x10;
			break;
		case 1:
			value |= 0x80000;
			break;
		default:
			return -EINVAL;
	}
	writel(value, pwm->regbase + PWM_CTL_REG);

	return 0;
}
EXPORT_SYMBOL(pwm_enable);

void pwm_disable(struct pwm_device *pwm)
{
	u32 value = readl(pwm->regbase + PWM_CTL_REG);
	switch (pwm->pwm_id)
	{
		case 0:
			value &= ~0x10;
			break;
		case 1:
			value &= ~0x80000;
			break;
		default:
			return;
	}
	writel(value, pwm->regbase + PWM_CTL_REG);

	return;
}
EXPORT_SYMBOL(pwm_disable);

static int __devinit sunxi_pwm_probe(struct platform_device *pdev)
{
	struct pwm_device *pwms;
	struct resource *res;
	void __iomem *base = NULL;
	int ret = 0, i;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
	{
		dev_err(&pdev->dev, "no memory resource\n");
		ret = -ENODEV;
		goto exit0;
	}

	res = request_mem_region(res->start, resource_size(res), pdev->name);
	if (!res)
	{
		dev_err(&pdev->dev, "request_mem_region failed\n");
		ret = -ENOMEM;
		goto exit0;
	}

	base = ioremap(res->start, resource_size(res));
	if (!base)
	{
		dev_err(&pdev->dev, "ioremap failed\n");
		ret = -ENOMEM;
		goto exit1;
	}

    pwms = kzalloc(PWMS_NR * sizeof(struct pwm_device), GFP_KERNEL);
	if (pwms == NULL)
	{
		dev_err(&pdev->dev, "alloc mem failed\n");
		ret = -ENOMEM;
		goto exit1;
	}

	for (i = 0; i < PWMS_NR; i++)
	{
		mutex_lock(&pwm_lock);
		pwms[i].use_count = 0;
		pwms[i].pwm_id = i;
		pwms[i].pdev = pdev;
		pwms[i].regbase = base;

		list_add_tail(&pwms[i].list, &pwm_list);
		mutex_unlock(&pwm_lock);
	}

	platform_set_drvdata(pdev, pwms);
	return 0;

exit1:
	release_mem_region(res->start, resource_size(res));
exit0:
	return ret;
}

static int __devexit sunxi_pwm_remove(struct platform_device *pdev)
{
	struct pwm_device *pwms;
	struct resource *res;
	int i;

	pwms = platform_get_drvdata(pdev);
	if (!pwms)
		return -ENODEV;

	mutex_lock(&pwm_lock);
	for (i = 0; i < PWMS_NR; i++)
		list_del(&pwms[i].list);
	mutex_unlock(&pwm_lock);

	iounmap(pwms[0].regbase);

	kfree(pwms);

	return 0;
}

static struct platform_driver sunxi_pwm_driver = {
	.driver		= {
		.name	= "sunxi-pwm",
		.owner	= THIS_MODULE,
	},
	.probe		= sunxi_pwm_probe,
	.remove		= __devexit_p(sunxi_pwm_remove),
};

static int __init sunxi_pwm_init(void)
{
	return platform_driver_register(&sunxi_pwm_driver);
}

arch_initcall(sunxi_pwm_init);

MODULE_LICENSE("GPL");
