/*
 *  linux/drivers/misc/mem.c
 *
 *  Copyright (C) 2013 XXX
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/dma-mapping.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <media/v4l2-ioctl.h>
#include <mach/includes.h>

#define DEFAULT_BUF_SIZE   (0x100000)

#define GS_MEM_MAJOR		220
#define SUN7I

enum {
    MEM_DEBUG_PROBE         = 1U << 0,
    MEM_DEBUG_ENTER         = 1U << 1,
    MEM_DEBUG_ALLOC         = 1U << 2,
    MEM_DEBUG_FREE          = 1U << 3,
    MEM_DEBUG_MMAP          = 1U << 4,
    MEM_DEBUG_6         = 1U << 5,
    MEM_DEBUG_7         = 1U << 6,
    MEM_DEBUG_8         = 1U << 7,
    MEM_DEBUG_9         = 1U << 8,
};

uint32_t mem_buf_size = DEFAULT_BUF_SIZE;
module_param_named(buf_size, mem_buf_size, uint, S_IRUGO);

uint32_t mem_debug_mask = 0;
module_param_named(debug_mask, mem_debug_mask, uint, S_IWUSR | S_IRUGO);

#define mem_debug(mask, x...) \
	do { \
		if (mem_debug_mask & mask) \
			printk("GS MEM: "x); \
	} while (0)

#ifdef SUN7I
unsigned long mem_start = GS_MEM_BASE;
unsigned long mem_size = GS_MEM_SIZE;
#endif

struct mem_info {
    unsigned char __iomem *kaddr;
    dma_addr_t paddr;
    unsigned long size;
    spinlock_t mem_lock;
};

struct mem_buf_info {
    unsigned char __iomem *cpu_addr;
    dma_addr_t dma_addr;
    size_t size;
    int in_used;
};

enum mem_buf_flag {
    MEM_BUF_RAW,
    MEM_BUF_ALLOCATED,
    MEM_BUF_CONNECTED,
};

struct mem_data {
    int index;              /* ID of buffer */
	size_t buf_size;		/* Size of buffer */
	char * buf_kaddr;		/* Start address of buffer */
	dma_addr_t buf_dmaaddr; /* Start physical address of buffer */
    enum mem_buf_flag flag; /* Flag indicate buffer state */
};

static struct mem_info *mem_info;
static struct mem_buf_info *mem_buf;
static int mem_buf_count;

static void mem_buf_init(struct mem_info *meminfo)
{
    int i;
    char *kaddr;
    dma_addr_t paddr;
    struct mem_buf_info *p_info;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    if (!meminfo->kaddr || !meminfo->paddr)
        return;

    mem_buf = kmalloc(sizeof(struct mem_buf_info) * mem_buf_count, GFP_KERNEL);
    if (mem_buf == NULL) {
        printk("GS MEM: mem_buf_init alloc mem_buf_info err!\n");
        return;
    }

    kaddr = (char *)meminfo->kaddr;
    paddr = (dma_addr_t)meminfo->paddr;
    p_info = mem_buf;

    for (i = 0; i < mem_buf_count; i++, p_info++) {
        p_info->cpu_addr = kaddr;
        p_info->dma_addr = paddr;
        p_info->size = mem_buf_size;
        p_info->in_used = 0;
        kaddr += mem_buf_size;
        paddr += mem_buf_size;
    }

    printk("GS MEM: %d buffers inited!\n", mem_buf_count);
    printk("GS MEM: kaddr %x, paddr %x\n", (unsigned int)kaddr, (unsigned int)paddr);
    return;
}

void *mem_dma_alloc(int *index, size_t *size, dma_addr_t *handle)
{
    int i;
    void *ret;
    struct mem_buf_info *p_info, *ret_info;
    unsigned long flags;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    if (mem_buf == NULL)
        return NULL;

    spin_lock_irqsave(&mem_info->mem_lock, flags);
    ret = ret_info = NULL;

    if (*size > mem_buf_size || *size <= 0) {
        printk("GS MEM: mem_buf_size %d, request %d\n", mem_buf_size, *size);
        ret = NULL;
        goto end;
    }

    p_info = mem_buf;
    for (i = 0; i < mem_buf_count; i++, p_info++) {
        if (p_info->in_used)
            continue;
        ret_info = p_info;
        *index = i;
        break;
    }

    if (ret_info) {
        *size = mem_buf_size;
        *handle = ret_info->dma_addr;
        ret = ret_info->cpu_addr;
        ret_info->in_used = 1;
        memset(ret_info->cpu_addr, 0, ret_info->size);
        mem_debug(MEM_DEBUG_ALLOC, "mem_dma_alloc kaddr %p, paddr %x, size %d\n",
                ret_info->cpu_addr, ret_info->dma_addr, *size);
    }

end:
    spin_unlock_irqrestore(&mem_info->mem_lock, flags);

    return ret;
}

void mem_dma_free(void *cpu_addr, dma_addr_t handle)
{
    int i;
    struct mem_buf_info *p_info;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    p_info = mem_buf;

    for (i = 0; i < mem_buf_count; i++, p_info++) {
        if (cpu_addr == p_info->cpu_addr && handle == p_info->dma_addr) {
            p_info->in_used = 0;
            mem_debug(MEM_DEBUG_FREE,"mem_dma_free OK: cpu_addr = 0x%p, dma_addr = 0x%x\n", 
                    cpu_addr, handle);
            return;
        }
        continue;
    }

    printk("GS MEM: mem_dma_free err cpu_addr = 0x%p, dma_addr = 0x%x\n", cpu_addr, handle);
    return;
}

//int mem_dma_ref(int index)
//{
//    if (index < 0 || index >= mem_buf_count) {
//        printk("%s: invalid index num %d\n", __func__, index);
//        return -EINVAL;
//    }
//}

int mem_dma_check_state(int index)
{
    int ret = 0;
    unsigned long flags;
    struct mem_buf_info *p_info;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    if (index < 0 || index >= mem_buf_count)
        return -EINVAL;

    p_info = mem_buf;

    spin_lock_irqsave(&mem_info->mem_lock, flags);

    ret = (p_info + index)->in_used;
    
    spin_unlock_irqrestore(&mem_info->mem_lock, flags);

    return ret;
}


static int gs_mem_REQBUF(struct mem_data *p_data, struct v4l2_buffer *buffer)
{
    int index;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    if (p_data->flag != MEM_BUF_RAW) {
        printk("GS MEM: QUERYBUF ERR, Invalid state %d\n", p_data->flag);
        return -EINVAL;
    }

    p_data->buf_kaddr = mem_dma_alloc(&index, &buffer->length, &p_data->buf_dmaaddr);
    if (p_data->buf_kaddr == NULL) {
        printk("GS MEM: ALLOC BUF ERR.\n");
        return -ENOMEM;
    }

    p_data->flag = MEM_BUF_ALLOCATED;
    p_data->buf_size = buffer->length;
    p_data->index = index;

    buffer->index = index;

	return 0;
}

static int gs_mem_QUERYBUF(struct mem_data *p_data, struct v4l2_buffer *buffer)
{
    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    buffer->length = mem_buf_size;
    return 0;
}

static long gs_mem_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct mem_data *p_data = file->private_data;
	struct v4l2_buffer buffer;
	int error = 0;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

	switch (cmd) {
	case VIDIOC_REQBUFS:
		error = copy_from_user(&buffer, (void *)arg, sizeof(buffer));
		if ( error )
			return error;
		error = gs_mem_REQBUF(p_data, &buffer);
		if ( ! error )
			error = copy_to_user((void *)arg, &buffer, sizeof(buffer));
		return error;

	case VIDIOC_QUERYBUF:
		error = copy_from_user(&buffer, (void *)arg, sizeof(buffer));
		if ( error )
			return error;
		error = gs_mem_QUERYBUF(p_data, &buffer);
		if ( ! error )
			error = copy_to_user((void *)arg, &buffer, sizeof(buffer));
		return error;

	default:
		return -EINVAL;
	}
	return error;
}

static ssize_t gs_mem_write(struct file *file, const char __user *buffer,
			size_t count, loff_t *ppos)
{
    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);
    return count;
}

#if 0
static unsigned int gs_mem_poll(struct file *file, poll_table *wait)
{
	struct mem_data * p_data = file->private_data;
	unsigned mask = 0;

    poll_wait(file, &p_data->read_wait, wait);
    poll_wait(file, &p_data->write_wait, wait);

    if ( p_data->input_ready )
        mask |= POLLOUT | POLLWRNORM;
    if ( p_data->output_ready )
        mask |= POLLIN | POLLRDNORM;
	return mask;
}
#endif

static int gs_mem_mmap(struct file *file, struct vm_area_struct * vma)
{
	//struct mem_data * p_data = file->private_data;
	unsigned long offset;
	int buf_index;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;

	offset = vma->vm_pgoff << PAGE_SHIFT;

    mem_debug(MEM_DEBUG_MMAP, "mmap: vm_pgoff %lx, vm_start %lx, vm_end %lx\n", 
                vma->vm_pgoff, vma->vm_start, vma->vm_end);
    mem_debug(MEM_DEBUG_MMAP, "mmap: vm_page_prot %x, vm_flags %lx\n", 
                vma->vm_page_prot, vma->vm_flags);

	buf_index = offset / mem_buf_size;
	//offset &= (0x00100000 - 1);

	if ( buf_index >= mem_buf_count || offset >= mem_info->size ) {
        printk("GS MEM: mmap invalid data index %d, offset 0x%lx\n", 
                    buf_index, offset);
		return -EINVAL;
    }

    if (mem_dma_check_state(buf_index) <= 0) {
        printk("GS MEM: mmap check err: buf %d not in used\n", 
                    buf_index);
        return -EINVAL;
    }

	/* Test map range */
	if ((vma->vm_end - vma->vm_start) > mem_buf_size) {
        printk("GS MEM: mmap range invalid\n");
		return -EINVAL;
    }

	if ((vma->vm_end - vma->vm_start + offset) > mem_info->size) {
        printk("GS MEM: mmap range and offset invalid\n");
		return -EINVAL;
    }

//	/* offset be DMA address */
	offset += (unsigned long)mem_info->paddr;
//	vma->vm_pgoff = offset >> PAGE_SHIFT;
	/* This is an IO map - tell maydump to skip this VMA */
	//vma->vm_flags |= VM_IO | VM_RESERVED;
	//fb_pgprotect(file, vma, offset);
	if (io_remap_pfn_range(vma, vma->vm_start, offset >> PAGE_SHIFT,
			     vma->vm_end - vma->vm_start, vma->vm_page_prot))
		return -EAGAIN;
	return 0;
}

static int gs_mem_close(struct inode *inode, struct file *file)
{
	struct mem_data * p_data = file->private_data;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    if (p_data->buf_kaddr)
        mem_dma_free (p_data->buf_kaddr, p_data->buf_dmaaddr);

	kfree (p_data);

	return 0;
}


static int gs_mem_open(struct inode *inode, struct file *file)
{
	struct mem_data * p_data;
	int error = 0;

    mem_debug(MEM_DEBUG_ENTER, "%s entered\n", __func__);

    p_data = kmalloc(sizeof(struct mem_data), GFP_KERNEL);

    if ( p_data == NULL )
        return -ENOMEM;

    memset(p_data, 0, sizeof(struct mem_data));

    p_data->flag = MEM_BUF_RAW;

    file->private_data = p_data;

    return error;
}

static struct device * gs_mem_dev;

static const struct file_operations gs_mem_fops = {
	.owner	= THIS_MODULE,
	.open	= gs_mem_open,
	.write = gs_mem_write,
	//.poll = gs_mem_poll,
	.release = gs_mem_close,
	.mmap = gs_mem_mmap,
	.unlocked_ioctl= gs_mem_ioctl,
};

static struct class * gs_mem_class;

static int gs_mem_probe(struct platform_device *pdev) 
{
#ifndef SUN7I
    struct resource *r;
#endif

	mem_info = kmalloc(sizeof(struct mem_info), GFP_KERNEL);
	if (mem_info == NULL) {
		printk("Unable to probe gs mem: kmalloc return MULL\n");
		return -1;
	}

#ifdef SUN7I
	mem_info->paddr = mem_start;
	mem_info->size = mem_size;
#else
    r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
    if (!r) {
        dev_err(&pdev->dev, "no memory resource given\n");
        return -EINVAL;
    }

	mem_info->paddr = r->start;
	mem_info->size = resource_size(r);
#endif

    if (!mem_info->paddr || mem_info->size <= 0) {
        printk("Unable to probe gs mem platform device\n");
        return -1;
    }

    //mem_info->kaddr = ioremap(mem_info->paddr, mem_info->size);
    mem_info->kaddr = __va(mem_info->paddr);
	if (!mem_info->kaddr) {
		printk("Unable to probe gs mem: ioremap return MULL\n");
        kfree(mem_info);
		return -1;
	}

	/* Create devices */
	gs_mem_dev = device_create(gs_mem_class, NULL,
			MKDEV(GS_MEM_MAJOR, 0), NULL, "gs_mem");

	spin_lock_init(&mem_info->mem_lock);

    mem_buf_count = mem_info->size / mem_buf_size;
	mem_buf_init(mem_info);

    mem_debug(MEM_DEBUG_PROBE, "v_mem_start = %p. p_mem_start = %x. mem_size = %lx\n", 
                mem_info->kaddr, mem_info->paddr, mem_info->size);

	return 0;
}

int gs_mem_remove(struct platform_device *pdev)
{
    device_unregister(gs_mem_dev);

//    if(mem_info->kaddr)
//        iounmap(mem_info->kaddr);

    if (mem_info != NULL)
        kfree(mem_info);

    mem_debug(MEM_DEBUG_PROBE, "module removed\n");

	return 0;
}

#ifdef CONFIG_PM
int gs_mem_suspend(struct platform_device *dev, pm_message_t state)
{
	return 0;
}

int gs_mem_resume(struct platform_device *dev)
{
	return 0;
}
#else
#define gs_mem_suspend	NULL
#define gs_mem_resume		NULL
#endif

static struct platform_driver gs_mem_driver = {
	.probe		= gs_mem_probe,
	.remove		= gs_mem_remove,
	.suspend	= gs_mem_suspend,
	.resume		= gs_mem_resume,
	.driver		= {
		.name	= "gs_mem",
	},
};

static int __devinit gs_mem_init(void)
{
	if (register_chrdev(GS_MEM_MAJOR, "gs-mem", &gs_mem_fops))
		printk("unable to get major %d for gs mem devs\n", GS_MEM_MAJOR);

	gs_mem_class = class_create(THIS_MODULE, "gs-mem");
	if (IS_ERR(gs_mem_class)) {
		printk(KERN_WARNING "Unable to create gs mem class; errno = %ld\n", PTR_ERR(gs_mem_class));
		gs_mem_class = NULL;
	}

	return platform_driver_register(&gs_mem_driver);
}

static void __exit gs_mem_exit(void)
{
	platform_driver_unregister(&gs_mem_driver);
	class_destroy(gs_mem_class);
	unregister_chrdev(GS_MEM_MAJOR, "gs-mem");
}

module_init(gs_mem_init);
module_exit(gs_mem_exit);

MODULE_AUTHOR("GS");
MODULE_DESCRIPTION("GS MEM");
MODULE_LICENSE("GPL");
