/*
 * Copyright (c) 2013 Grandstream Networks, Inc.
 * By Fanan Chen <fnchen@grandstream.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option)any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <linux/gpio.h>
#include <linux/platform_device.h>
#include <mach/gs_plat.h>

struct gs_hook_priv {
	int hook_status;
	int gpio;
	unsigned long int_handle;
	struct timer_list timer;
};

#define	HOOK_DELAY	100

static struct gs_hook_priv *g_hook;

static int proc_key_hook(char *page, char **start, off_t off,
		int count, int *eof, void *data)
{
	char *status = (char *)page;
	char *str;
	int len = 0;
	str = g_hook->hook_status ? "up" : "down";
	len = strlen(str);
	memcpy(status, str, len);
	if (len <= off+count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

static void gs_hook_timer(unsigned long data)
{
	struct gs_hook_priv *hook = (struct gs_hook_priv *)data;
	int gpio = hook->gpio;
	int val = 0;

	val = gpio_get_value(gpio);
	if (val != hook->hook_status)
		hook->hook_status = val;

	sw_gpio_eint_set_enable(gpio, 1);
}

static int hook_interrupt(void *data)
{
	struct gs_hook_priv *hook = (struct gs_hook_priv *)data;

	sw_gpio_eint_set_enable(hook->gpio, 0);
	mod_timer(&hook->timer, jiffies + msecs_to_jiffies(HOOK_DELAY));

	return 0;
}

static int __devinit gs_hook_probe(struct platform_device *pdev)
{
	struct gs_hook_pdata *pdata = pdev->dev.platform_data;
	struct gs_hook_priv *hook = NULL;
	struct proc_dir_entry *ent = NULL;

	hook = kzalloc(sizeof(struct gs_hook_priv), GFP_KERNEL);
	if (!hook)
		return -ENOMEM;

	init_timer(&hook->timer);
	hook->timer.function = gs_hook_timer;
	hook->timer.data = (unsigned long)hook;
	hook->gpio = pdata->gpio;

	hook->int_handle = sw_gpio_irq_request(pdata->gpio,
			TRIG_EDGE_DOUBLE, (peint_handle)hook_interrupt, hook);
	if (!hook->int_handle) {
		pr_err("gs_hook:request irq failed\n");
		return -EINVAL;
	}

	hook->hook_status = gpio_get_value(pdata->gpio);
	g_hook = hook;

	ent = create_proc_entry("key_hook", S_IFREG|S_IRUGO, NULL);
	if (!ent)
		return -ENODEV;
	ent->write_proc = NULL;
	ent->read_proc = proc_key_hook;

	return 0;
}

static int __devexit gs_hook_remove(struct platform_device *pdev)
{

	sw_gpio_irq_free(g_hook->int_handle);
	return 0;
}

static struct platform_driver gs_hook_driver = {
	.probe		= gs_hook_probe,
	.remove		= __devexit_p(gs_hook_remove),
	.driver		= {
		.name	= "hook-switch",
		.owner	= THIS_MODULE,
	},
};

static int __init gs_hook_init(void)
{
	return platform_driver_register(&gs_hook_driver);
}

static void __exit gs_hook_exit(void)
{
	platform_driver_unregister(&gs_hook_driver);
}

module_init(gs_hook_init);
module_exit(gs_hook_exit);

MODULE_AUTHOR("Fanan Chen <fnchen@grandstream.com>");
MODULE_DESCRIPTION("Grandstream Hook Switch driver");
MODULE_LICENSE("GPL");
