/*
 * (C) Copyright 2007-2011
 * Allwinner Technology Co., Ltd. <www.allwinnertech.com>
 * Aaron <leafy.myeh@allwinnertech.com>
 *
 * i2c driver for allwinner sunxi platform.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <asm/arch/gpio.h>
#include <i2c.h>
#include "sun7i_i2c.h"

#ifdef CONFIG_HARD_I2C

#define SUN7I_I2C_OK      0
#define SUN7I_I2C_FAIL   -1
#define SUN7I_I2C_RETRY  -2
#define SUN7I_I2C_SFAIL  -3  /* start fail */
#define SUN7I_I2C_TFAIL  -4  /* stop  fail */
#define SUN7I_I2C_END   1 

#undef SUN7I_I2C_DEBUG
#ifdef SUN7I_I2C_DEBUG
#define I2C_DEBUG(fmt, args...) printf(fmt, ##args)
#define pr_debug(fmt, args...) printf(fmt, ##args)
#else
#define I2C_DEBUG(...) do { } while(0)
#define pr_debug(...) do { } while(0)
#endif

#define I2C_ERR(fmt, args...) printf(fmt, ##args)

static struct sun7i_i2c i2c0 =
{
	.bus_num = 0,
	.bus_freq = TWI0_TRANSFER_SPEED,
	.base_addr = (void*)SW_PA_TWI0_IO_BASE,	
};

static struct sun7i_i2c i2c1 =
{
	.bus_num = 1,
	.bus_freq = TWI1_TRANSFER_SPEED,
	.base_addr = (void*)SW_PA_TWI1_IO_BASE,	
};

static struct sun7i_i2c i2c2 =
{
	.bus_num = 2,
	.bus_freq = TWI2_TRANSFER_SPEED,
	.base_addr = (void*)SW_PA_TWI2_IO_BASE,	
};

static inline void twi_clear_irq_flag(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val &= ~TWI_CTL_INTFLG;//0x 1111_0111
	writel(reg_val ,base_addr + TWI_CTL_REG);
	/* read two more times to make sure that interrupt flag does really be cleared */
	{
		unsigned int temp;
		temp = readl(base_addr + TWI_CTL_REG);
		temp |= readl(base_addr + TWI_CTL_REG);
	}
}

/* get data first, then clear flag */
static inline void twi_get_byte(void *base_addr, unsigned char  *buffer)
{
	*buffer = (unsigned char)( TWI_DATA_MASK & readl(base_addr + TWI_DATA_REG) );
	twi_clear_irq_flag(base_addr);
}

/* only get data, we will clear the flag when stop */
static inline void twi_get_last_byte(void *base_addr, unsigned char  *buffer)
{
	*buffer = (unsigned char)( TWI_DATA_MASK & readl(base_addr + TWI_DATA_REG) );
}

/* write data and clear irq flag to trigger send flow */
static inline void twi_put_byte(void *base_addr, const unsigned char *buffer)
{
	writel((unsigned int)*buffer, base_addr + TWI_DATA_REG);
	twi_clear_irq_flag(base_addr);
}

static inline void twi_enable_irq(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);

	/*
	 * 1 when enable irq for next operation, set intflag to 1 to prevent to clear it by a mistake
	 *   (intflag bit is write-0-to-clear bit)
	 * 2 Similarly, mask START bit and STOP bit to prevent to set it twice by a mistake
	 *   (START bit and STOP bit are self-clear-to-0 bits)
	 */
	reg_val |= (TWI_CTL_INTEN | TWI_CTL_INTFLG);
	reg_val &= ~(TWI_CTL_STA | TWI_CTL_STP);
	writel(reg_val, base_addr + TWI_CTL_REG);
}

static inline void twi_disable_irq(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val &= ~TWI_CTL_INTEN;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

static inline void twi_disable_bus(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val &= ~TWI_CTL_BUSEN;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

static inline void twi_enable_bus(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val |= TWI_CTL_BUSEN;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

/* trigger start signal, the start bit will be cleared automatically */
static inline void twi_set_start(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val |= TWI_CTL_STA;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

/* get start bit status, poll if start signal is sent */
static inline unsigned int twi_get_start(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val >>= 5;
	return reg_val & 1;
}

/* trigger stop signal, the stop bit will be cleared automatically */
static inline void twi_set_stop(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val |= TWI_CTL_STP;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

/* get stop bit status, poll if stop signal is sent */
static inline unsigned int twi_get_stop(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val >>= 4;
	return reg_val & 1;
}

static inline void twi_disable_ack(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val &= ~TWI_CTL_ACK;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

/* when sending ack or nack, it will send ack automatically */
static inline void twi_enable_ack(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	reg_val |= TWI_CTL_ACK;
	writel(reg_val, base_addr + TWI_CTL_REG);
}

/* get the interrupt flag */
static inline unsigned int twi_query_irq_flag(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CTL_REG);
	return (reg_val & TWI_CTL_INTFLG);//0x 0000_1000
}

/* get interrupt status */
static inline unsigned int twi_query_irq_status(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_STAT_REG);
	return (reg_val & TWI_STAT_MASK);
}

/* set twi clock
 *
 * clk_n: clock divider factor n
 * clk_m: clock divider factor m
 */
static inline void twi_clk_write_reg(unsigned int clk_n, unsigned int clk_m, void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_CLK_REG);
	pr_debug("%s: clk_n = %d, clk_m = %d\n", __FUNCTION__, clk_n, clk_m);
	reg_val &= ~(TWI_CLK_DIV_M | TWI_CLK_DIV_N);
	reg_val |= ( clk_n |(clk_m << 3) );
	writel(reg_val, base_addr + TWI_CLK_REG);
}

/*
* Fin is APB CLOCK INPUT;
* Fsample = F0 = Fin/2^CLK_N;
* F1 = F0/(CLK_M+1);
* Foscl = F1/10 = Fin/(2^CLK_N * (CLK_M+1)*10);
* Foscl is clock SCL;100KHz or 400KHz
*
* clk_in: apb clk clock
* sclk_req: freqence to set in HZ
*/
static void twi_set_clock(unsigned int clk_in, unsigned int sclk_req, void *base_addr)
{
	unsigned int clk_m = 0;
	unsigned int clk_n = 0;
	unsigned int _2_pow_clk_n = 1;
	unsigned int src_clk      = clk_in/10;
	unsigned int divider      = src_clk/sclk_req;  // 400khz or 100khz
	unsigned int sclk_real    = 0;      // the real clock frequency

	if (divider == 0) {
		clk_m = 1;
		goto set_clk;
	}

	/* search clk_n and clk_m,from large to small value so that can quickly find suitable m & n. */
	while (clk_n < 8) { // 3bits max value is 8
		/* (m+1)*2^n = divider -->m = divider/2^n -1 */
		clk_m = (divider/_2_pow_clk_n) - 1;
		/* clk_m = (divider >> (_2_pow_clk_n>>1))-1 */
		while (clk_m < 16) { /* 4bits max value is 16 */
			sclk_real = src_clk/(clk_m + 1)/_2_pow_clk_n;  /* src_clk/((m+1)*2^n) */
			if (sclk_real <= sclk_req) {
				goto set_clk;
			}
			else {
				clk_m++;
			}
		}
		clk_n++;
		_2_pow_clk_n *= 2; /* mutilple by 2 */
	}

set_clk:
	twi_clk_write_reg(clk_n, clk_m, base_addr);

	return;
}

/* soft reset twi */
static inline void twi_soft_reset(void *base_addr)
{
	unsigned int reg_val = readl(base_addr + TWI_SRST_REG);
	reg_val |= TWI_SRST_SRST;
	writel(reg_val, base_addr + TWI_SRST_REG);
}

/* Enhanced Feature Register */
static inline void twi_set_efr(void *base_addr, unsigned int efr)
{
	unsigned int reg_val = readl(base_addr + TWI_EFR_REG);

	reg_val &= ~TWI_EFR_MASK;
	efr     &= TWI_EFR_MASK;
	reg_val |= efr;
	writel(reg_val, base_addr + TWI_EFR_REG);
}

/* function  */
static int twi_start(void *base_addr, int bus_num)
{
	unsigned int timeout = 0xff;

	twi_set_start(base_addr);
	while((1 == twi_get_start(base_addr))&&(--timeout));
	if(timeout == 0) {
		I2C_ERR("[i2c%d] START can't sendout!\n", bus_num);
		return SUN7I_I2C_FAIL;
	}

	return SUN7I_I2C_OK;
}

static int twi_restart(void  *base_addr, int bus_num)
{
	unsigned int timeout = 0xff;
	twi_set_start(base_addr);
	twi_clear_irq_flag(base_addr);
	while((1 == twi_get_start(base_addr))&&(--timeout));
	if(timeout == 0) {
		I2C_ERR("[i2c%d] Restart can't sendout!\n", bus_num);
		return SUN7I_I2C_FAIL;
	}
	return SUN7I_I2C_OK;
}

static int twi_stop(void *base_addr, int bus_num)
{
	unsigned int timeout = 0xff;

	twi_set_stop(base_addr);
	twi_clear_irq_flag(base_addr);

	twi_get_stop(base_addr);/* it must delay 1 nop to check stop bit */
	while(( 1 == twi_get_stop(base_addr))&& (--timeout));
	if(timeout == 0) {
		I2C_ERR("[i2c%d] STOP can't sendout!\n", bus_num);
		return SUN7I_I2C_TFAIL;
	}

	timeout = 0xffff;
	while((TWI_STAT_IDLE != readl(base_addr + TWI_STAT_REG))&&(--timeout));
	if(timeout == 0)
	{
		I2C_ERR("[i2c%d] i2c state isn't idle(0xf8)\n", bus_num);
		return SUN7I_I2C_TFAIL;
	}

	timeout = 0xff;
	while((TWI_LCR_IDLE_STATUS != readl(base_addr + TWI_LCR_REG))&&(--timeout));
	if(timeout == 0) {
		I2C_ERR("[i2c%d] i2c lcr isn't idle(0x3a)\n", bus_num);
		return SUN7I_I2C_TFAIL;
	}

	return SUN7I_I2C_OK;
}

/* get SDA state */
static unsigned int twi_get_sda(void *base_addr)
{
        unsigned int status = 0;
        status = TWI_LCR_SDA_STATE_MASK & readl(base_addr + TWI_LCR_REG);
        status >>= 4;
        return  (status&0x1);
}

/* set SCL level(high/low), only when SCL enable */
static void twi_set_scl(void *base_addr, unsigned int hi_lo)
{
        unsigned int reg_val = readl(base_addr + TWI_LCR_REG);
        reg_val &= ~TWI_LCR_SCL_CTL;
        hi_lo   &= 0x01;
        reg_val |= (hi_lo<<3);
        writel(reg_val, base_addr + TWI_LCR_REG);
}

/* enable SDA or SCL */
static void twi_enable_lcr(void *base_addr, unsigned int sda_scl)
{
        unsigned int reg_val = readl(base_addr + TWI_LCR_REG);
        sda_scl &= 0x01;
        
        if(sda_scl) {
                reg_val |= TWI_LCR_SCL_EN;/* enable scl line control */
        } else {
                reg_val |= TWI_LCR_SDA_EN;/* enable sda line control */
        }
        writel(reg_val, base_addr + TWI_LCR_REG);
}

/* disable SDA or SCL */
static void twi_disable_lcr(void *base_addr, unsigned int sda_scl)
{
        unsigned int reg_val = readl(base_addr + TWI_LCR_REG);
        sda_scl &= 0x01;
        
        if(sda_scl) {
                reg_val &= ~TWI_LCR_SCL_EN;/* disable scl line control */
        } else {
                reg_val &= ~TWI_LCR_SDA_EN;/* disable sda line control */
        }
        writel(reg_val, base_addr + TWI_LCR_REG);
}

/* send 9 clock to release sda */
static int twi_send_clk_9pulse(void *base_addr, int bus_num)
{
        int twi_scl = 1;
        int low = 0;
        int high = 1;
        int cycle = 0;

        /* enable scl control */
        twi_enable_lcr(base_addr, twi_scl);

        while(cycle < 10) {
                if( twi_get_sda(base_addr)
                && twi_get_sda(base_addr)
                && twi_get_sda(base_addr) ) {
                        break;
                }
                /* twi_scl -> low */
                twi_set_scl(base_addr, low);
                udelay(1000);

                /* twi_scl -> high */
                twi_set_scl(base_addr, high);
                udelay(1000);
                cycle++;
        }

        if(twi_get_sda(base_addr)){
                twi_disable_lcr(base_addr, twi_scl);
                return SUN7I_I2C_OK;
        } else {
        I2C_ERR("[i2c%d] SDA is still Stuck Low, failed. \n", bus_num);
                twi_disable_lcr(base_addr, twi_scl);
                return SUN7I_I2C_FAIL;
        }
}

static int sun7i_i2c_clk_init(struct sun7i_i2c *i2c)
{
	unsigned int apb_clk = 24000000;

	/* enable twi bus */
	twi_enable_bus(i2c->base_addr);
	twi_set_clock(apb_clk, i2c->bus_freq, i2c->base_addr);

	return 0;
}

/*
 * cmd_type is 0 for write, 1 for read.
 *
 * addr_len can take any value from 0-255, it is only limited
 * by the char, we could make it larger if needed. If it is
 * 0 we skip the address write cycle.
 */
static int i2c_transfer(unsigned char cmd_type,
		 unsigned char chip,
		 unsigned char addr[],
		 unsigned char addr_len,
		 unsigned char data[], unsigned short data_len)
{	
	return 0;
}

/*
****************************************************************************************************
*
*  FunctionName:           sun7i_i2c_addr_byte
*
*  Description:
*            slaveַ7bitȫϢ10bitĵһֵַⲿӿڵãڲʵ֡
*         7bits addr: 7-1bits addr+0 bit r/w
*         10bits addr: 1111_11xx_xxxx_xxxx-->1111_0xx_rw,xxxx_xxxx
*         send the 7 bits addr,or the first part of 10 bits addr
*  Parameters:
*
*
*  Return value:
*           
*  Notes:
*
****************************************************************************************************
*/
static void sun7i_i2c_addr_byte(struct sun7i_i2c *i2c)
{
	unsigned char addr = 0;
	unsigned char tmp  = 0;

	if(i2c->msg[i2c->msg_idx].flags & I2C_M_TEN) {
		/* 0111_10xx,ten bits address--9:8bits */
		tmp = 0x78 | ( ( (i2c->msg[i2c->msg_idx].addr)>>8 ) & 0x03);
		addr = tmp << 1;//1111_0xx0
		/* how about the second part of ten bits addr? Answer: deal at twi_core_process() */
	} else {
		/* 7-1bits addr, xxxx_xxx0 */
		addr = (i2c->msg[i2c->msg_idx].addr & 0x7f) << 1;
	}

	/* read, default value is write */
	if (i2c->msg[i2c->msg_idx].flags & I2C_M_RD) {
		addr |= 1;
	}

	if(i2c->msg[i2c->msg_idx].flags & I2C_M_TEN) {
		I2C_DEBUG("[i2c%d] first part of 10bits = 0x%x\n", i2c->bus_num, addr);
	}
	I2C_DEBUG("[i2c%d] 7bits+r/w = 0x%x\n", i2c->bus_num, addr);

	/* send 7bits+r/w or the first part of 10bits */
	twi_put_byte(i2c->base_addr, &addr);
}

static int sun7i_i2c_xfer_complete(struct sun7i_i2c *i2c, int code)
{
	int ret = SUN7I_I2C_OK;

	i2c->msg     = NULL;
	i2c->msg_num = 0;
	i2c->msg_ptr = 0;
	i2c->status  = I2C_XFER_IDLE;

	/* i2c->msg_idx  store the information */
	if(code == SUN7I_I2C_FAIL) {
		I2C_ERR("[i2c%d] Maybe Logic Error, debug it!\n", i2c->bus_num);
		i2c->msg_idx = code;
		ret = SUN7I_I2C_FAIL;
	}
	else if(code != SUN7I_I2C_OK) {
		i2c->msg_idx = code;
		ret = SUN7I_I2C_FAIL;
	}

	return ret;
}

static int sun7i_i2c_core_process(struct sun7i_i2c *i2c)
{
	void *base_addr = i2c->base_addr;
	int  ret        = SUN7I_I2C_OK;
	int  err_code   = 0;
	unsigned char  state = 0;
	unsigned char  tmp   = 0;

	state = twi_query_irq_status(base_addr);

	//I2C_DEBUG("[i2c%d][slave address = (0x%x), state = (0x%x)]\n", i2c->bus_num, i2c->msg->addr, state);

	if(i2c->msg == NULL) {
		I2C_ERR("[i2c%d] i2c message is NULL, err_code = 0xfe\n", i2c->bus_num);
		err_code = 0xfe;
		goto msg_null;
	}

	switch(state) {
	case 0xf8: /* On reset or stop the bus is idle, use only at poll method */
		err_code = 0xf8;
		goto err_out;
	case 0x08: /* A START condition has been transmitted */
	case 0x10: /* A repeated start condition has been transmitted */
		sun7i_i2c_addr_byte(i2c);/* send slave address */
		break;
	case 0xd8: /* second addr has transmitted, ACK not received!    */
	case 0x20: /* SLA+W has been transmitted; NOT ACK has been received */
		err_code = 0x20;
		goto err_out;
	case 0x18: /* SLA+W has been transmitted; ACK has been received */
		/* if any, send second part of 10 bits addr */
		if(i2c->msg[i2c->msg_idx].flags & I2C_M_TEN) {
			tmp = i2c->msg[i2c->msg_idx].addr & 0xff;  /* the remaining 8 bits of address */
			twi_put_byte(base_addr, &tmp); /* case 0xd0: */
			break;
		}
		/* for 7 bit addr, then directly send data byte--case 0xd0:  */
	case 0xd0: /* second addr has transmitted,ACK received!     */
	case 0x28: /* Data byte in DATA REG has been transmitted; ACK has been received */
		/* after send register address then START send write data  */
		if(i2c->msg_ptr < i2c->msg[i2c->msg_idx].len) {
			twi_put_byte(base_addr, &(i2c->msg[i2c->msg_idx].buf[i2c->msg_ptr]));
			i2c->msg_ptr++;
			break;
		}

		i2c->msg_idx++; /* the other msg */
		i2c->msg_ptr = 0;
		
		if (i2c->msg_idx == i2c->msg_num) {
			err_code = SUN7I_I2C_OK;/* Success,wakeup */
			goto ok_out;
		} else if(i2c->msg_idx < i2c->msg_num) {/* for restart pattern */
			ret = twi_restart(base_addr, i2c->bus_num);/* read spec, two msgs */
			
			if(ret == SUN7I_I2C_FAIL) {
				err_code = SUN7I_I2C_SFAIL;
				goto err_out;/* START can't sendout */
			}
		} else {
			err_code = SUN7I_I2C_FAIL;
			goto err_out;
		}
		break;
	case 0x30: /* Data byte in I2CDAT has been transmitted; NOT ACK has been received */
		err_code = 0x30;//err,wakeup the thread
		goto err_out;
	case 0x38: /* Arbitration lost during SLA+W, SLA+R or data bytes */
		err_code = 0x38;//err,wakeup the thread
		goto err_out;
	case 0x40: /* SLA+R has been transmitted; ACK has been received */
		/* with Restart,needn't to send second part of 10 bits addr,refer-"I2C-SPEC v2.1" */
		/* enable A_ACK need it(receive data len) more than 1. */
		if(i2c->msg[i2c->msg_idx].len > 1) {
			/* send register addr complete,then enable the A_ACK and get ready for receiving data */
			twi_enable_ack(base_addr);
			twi_clear_irq_flag(base_addr);/* jump to case 0x50 */
		} else if(i2c->msg[i2c->msg_idx].len == 1) {
			twi_clear_irq_flag(base_addr);/* jump to case 0x58 */
		}
		break;
	case 0x48: /* SLA+R has been transmitted; NOT ACK has been received */
		err_code = 0x48;//err,wakeup the thread
		goto err_out;
	case 0x50: /* Data bytes has been received; ACK has been transmitted */
		/* receive first data byte */
		if (i2c->msg_ptr < i2c->msg[i2c->msg_idx].len) {
			/* more than 2 bytes, the last byte need not to send ACK */
			if( (i2c->msg_ptr + 2) == i2c->msg[i2c->msg_idx].len ) {
				twi_disable_ack(base_addr);/* last byte no ACK */
			}
			/* get data then clear flag,then next data comming */
			twi_get_byte(base_addr, &i2c->msg[i2c->msg_idx].buf[i2c->msg_ptr]);
			i2c->msg_ptr++;
			break;
		}
		/* err process, the last byte should be @case 0x58 */
		err_code = SUN7I_I2C_FAIL;/* err, wakeup */
		goto err_out;
	case 0x58: /* Data byte has been received; NOT ACK has been transmitted */
		/* received the last byte  */
		if ( i2c->msg_ptr == i2c->msg[i2c->msg_idx].len - 1 ) {
			twi_get_last_byte(base_addr, &i2c->msg[i2c->msg_idx].buf[i2c->msg_ptr]);
			i2c->msg_idx++;
			i2c->msg_ptr = 0;
			if (i2c->msg_idx == i2c->msg_num) {
				err_code = SUN7I_I2C_OK; // succeed,wakeup the thread
				goto ok_out;
			} else if(i2c->msg_idx < i2c->msg_num) {
				/* repeat start */
				ret = twi_restart(base_addr, i2c->bus_num);
				if(ret == SUN7I_I2C_FAIL) {/* START fail */
					err_code = SUN7I_I2C_SFAIL;
					goto err_out;
				}
				break;
			}
		} else {
			err_code = 0x58;
			goto err_out;
		}
	case 0x00: /* Bus error during master or slave mode due to illegal level condition */
		err_code = 0xff;
		goto err_out;
	default:
		err_code = state;
		goto err_out;
	}
	i2c->debug_state = state;/* just for debug */
	return ret;

ok_out:
err_out:
	if(SUN7I_I2C_TFAIL == twi_stop(base_addr, i2c->bus_num)) {
		I2C_ERR("[i2c%d] STOP failed!\n", i2c->bus_num);
	}

msg_null:
	ret = sun7i_i2c_xfer_complete(i2c, err_code);/* wake up */
	i2c->debug_state = state;/* just for debug */
	return ret;
}

static int sun7i_i2c_handler(struct sun7i_i2c *i2c)
{
	int ret = SUN7I_I2C_FAIL;

	if(!twi_query_irq_flag(i2c->base_addr)) {
		udelay(100);
		//I2C_ERR("unknown interrupt!\n");
		return ret;
	}

	/* disable irq */
	twi_disable_irq(i2c->base_addr);

	/* twi core process */
	ret = sun7i_i2c_core_process(i2c);

	/* enable irq only when twi is transfering, otherwise disable irq */
	if(i2c->status != I2C_XFER_IDLE) {
		twi_enable_irq(i2c->base_addr);
	}

	return SUN7I_I2C_OK;
}

static int sun7i_i2c_do_xfer(struct sun7i_i2c *i2c, struct i2c_msg *msgs, int num)
{
	int ret = SUN7I_I2C_FAIL;

	twi_soft_reset(i2c->base_addr);
	udelay(100);

	/* test the bus is free,already protect by the semaphore at DEV layer */
	while( TWI_STAT_IDLE != twi_query_irq_status(i2c->base_addr)&&
	       TWI_STAT_BUS_ERR != twi_query_irq_status(i2c->base_addr) &&
	       TWI_STAT_ARBLOST_SLAR_ACK != twi_query_irq_status(i2c->base_addr) ) {
		pr_debug("[i2c%d] bus is busy, status = %x\n", i2c->bus_num, twi_query_irq_status(i2c->base_addr));
                if( SUN7I_I2C_OK == twi_send_clk_9pulse(i2c->base_addr, i2c->bus_num) ) {
                        break;
                } else {
                        ret = SUN7I_I2C_RETRY;
                        goto out;
                }
	}

	/* may conflict with xfer_complete */
	i2c->msg     = msgs;
	i2c->msg_num = num;
	i2c->msg_ptr = 0;
	i2c->msg_idx = 0;
	i2c->status  = I2C_XFER_START;
       twi_enable_irq(i2c->base_addr);  /* enable irq */
	twi_disable_ack(i2c->base_addr); /* disabe ACK */
	twi_set_efr(i2c->base_addr, 0);  /* set the special function register,default:0. */

//	for(i =0 ; i < num; i++){
//		for(j = 0; j < msgs->len; j++){
//			I2C_DEBUG("baddr = 0x%x \n",msgs->addr);
//			I2C_DEBUG("data = 0x%x \n", msgs->buf[j]);
//		}
//		I2C_DEBUG("\n\n");
//	}

	/* START signal, needn't clear int flag */
	ret = twi_start(i2c->base_addr, i2c->bus_num);
	if(ret == SUN7I_I2C_FAIL) {
		twi_soft_reset(i2c->base_addr);
		twi_disable_irq(i2c->base_addr);  /* disable irq */
		i2c->status  = I2C_XFER_IDLE;
		ret = SUN7I_I2C_RETRY;
		goto out;
	}

	i2c->status  = I2C_XFER_RUNNING;

	while(i2c->msg_num)
		ret = sun7i_i2c_handler(i2c);
	
	/* return code,if(msg_idx == num) succeed */
	ret = i2c->msg_idx;

	if (ret != num) {
		I2C_ERR("[i2c%d] incomplete xfer (status: 0x%x, dev addr: 0x%x)\n", i2c->bus_num, ret, msgs->addr);
		ret = -ECOMM;
	}
out:
	return ret;
}

void i2c0_init(int speed, int slaveadd)
{
	int ret = 0;
	struct sun7i_i2c*  i2c = &i2c0;

	unsigned int reg_val =  *((volatile unsigned long  *)(0x01C2800C));
	//reg_val |= TWI_CTL_BUSEN;
	//*((volatile unsigned long  *)(0x01C2800C)) = (unsigned long)(reg_val);
	
	pr_debug("twi_enable_bus1 0x%x, 0x%x \n", *((volatile unsigned long  *)(0x01C2800C)), reg_val);

	//reg_val =  *((volatile unsigned long  *)(0x01C2800C));
	//reg_val |= TWI_CTL_BUSEN;
	//*((volatile unsigned long  *)(0x01C2800C)) = (unsigned long)(reg_val);
	
	pr_debug("twi_enable_bus2 0x%x \n", *((volatile unsigned long  *)(0x01C2800C)));
	
	
	twi_soft_reset(i2c->base_addr);
	if(sun7i_i2c_clk_init(i2c)) {
		pr_debug("[i2c%d] init i2c clock failed!\n", i2c->bus_num);
		return;
	}

	twi_soft_reset(i2c->base_addr);

}

void i2c_init(int speed, int slaveadd)
{
	int ret = 0;
	struct sun7i_i2c*  i2c = &i2c1;

	twi_soft_reset(i2c->base_addr);
	if(sun7i_i2c_clk_init(i2c)) {
		pr_debug("[i2c%d] init i2c clock failed!\n", i2c->bus_num);
		return;
	}

	twi_soft_reset(i2c->base_addr);
	
	pr_debug("**********start************\n");
	pr_debug("0x%x \n",readl(i2c->base_addr + 0x0c));
	pr_debug("0x%x \n",readl(i2c->base_addr + 0x10));
	pr_debug("0x%x \n",readl(i2c->base_addr + 0x14));
	pr_debug("0x%x \n",readl(i2c->base_addr + 0x18));
	pr_debug("0x%x \n",readl(i2c->base_addr + 0x1c));
	pr_debug("**********end************\n");

}

void i2c2_init(int speed, int slaveadd)
{
	int ret = 0;
	struct sun7i_i2c*  i2c = &i2c2;

	twi_soft_reset(i2c->base_addr);
	if(sun7i_i2c_clk_init(i2c)) {
		pr_debug("[i2c%d] init i2c clock failed!\n", i2c->bus_num);
		return;
	}

	twi_soft_reset(i2c->base_addr);

}


int i2c_probe(uchar chip)
{
	/*
	 * What is needed is to send the chip address and verify that the
	 * address was <ACK>ed (i.e. there was a chip at that address which
	 * drove the data line low).
	 */
	u8 buf[3];
	int ret = -1;

	struct sun7i_i2c*  i2c = &i2c1;
	buf[0] = 0;
	buf[1] = 0;

	struct i2c_msg msg[] = {
		{
			.addr	= chip,
			.flags	= 0,
			.len	= 2,
			.buf	= buf,
		},
	};

	ret = sun7i_i2c_do_xfer(i2c, msg, 1);
	if (ret < 0)
		pr_debug("%s i2c xfer error: %d\n", __func__, ret);

	return ret;
	
}

int i2c0_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	int i, ret = -1;
	struct sun7i_i2c*  i2c = &i2c0;
	unsigned char buf[4]={0};

	if(alen>2)
		return -1;
	
	for(i=0;i<alen;i++)
       {
		buf[alen-i-1] = (addr >> (i*8)) & 0xFF;
       }
	
	struct i2c_msg msgs[] = {
		{
			.addr	= chip,
			.flags	= 0,
			.len	= alen,
			.buf	= buf,
		},
		{
			.addr	= chip,
			.flags	= I2C_M_RD,
			.len	= len,
			.buf	= buffer,
		},
	};

	ret = sun7i_i2c_do_xfer(i2c, msgs, 1);
	if (ret < 0)
		pr_debug("[i2c%d] %s i2c0 xfer error: %d\n", i2c->bus_num, __func__, ret);
	
	ret = sun7i_i2c_do_xfer(i2c, &msgs[1], 1);
	if (ret < 0)
		pr_debug("[i2c%d] %s i2c0 xfer error: %d\n", i2c->bus_num, __func__, ret);
	
	return ret;
	
}

int i2c_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	int i, ret = -1;
	struct sun7i_i2c*  i2c = &i2c1;
	unsigned char buf[4]={0};

	if(alen>2)
		return -1;

	#if 0
	for(i=0;i<alen;i++)
       {
		buf[alen-i-1] = (addr >> (i*8)) & 0xFF;
       }
	#endif
	
	struct i2c_msg msgs[] = {
		{
			.addr	= chip,
			.flags	= 0,
			.len	= 1,
			.buf	= buf,
		},
		{
			.addr	= chip,
			.flags	= I2C_M_RD,
			.len	= len,
			.buf	= buffer,
		},
	};

	ret = sun7i_i2c_do_xfer(i2c, msgs, 2);
	if (ret < 0)
		pr_debug("%s i2c xfer error: %d\n", __func__, ret);

	return ret;
	
}

int i2c2_read(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	int i, ret = -1;
	struct sun7i_i2c*  i2c = &i2c2;
	unsigned char buf[4]={0};

	if(alen>2)
		return -1;
	
	for(i=0;i<alen;i++)
       {
		buf[alen-i-1] = (addr >> (i*8)) & 0xFF;
       }
	
	struct i2c_msg msgs[] = {
		{
			.addr	= chip,
			.flags	= 0,
			.len	= alen,
			.buf	= buf,
		},
		{
			.addr	= chip,
			.flags	= I2C_M_RD,
			.len	= len,
			.buf	= buffer,
		},
	};

	ret = sun7i_i2c_do_xfer(i2c, msgs, 2);
	if (ret < 0)
		pr_debug("%s i2c xfer error: %d\n", __func__, ret);

	return ret;
	
}

int i2c_write(uchar chip, uint addr, int alen, uchar *buffer, int len)
{
	int i, ret = -1;
	struct sun7i_i2c*  i2c = &i2c1;
       unsigned char *data = malloc(alen+len);

	if(alen>2)
		return -1;

       for(i=0;i<alen;i++)
       {
		data[alen-i-1] = (addr >> (i*8)) & 0xFF;
       }
	memcpy(data+alen, buffer, len);
		
	struct i2c_msg msg[] = {
		{
			.addr	= chip,
			.flags	= 0,
			.len	= len+alen,
			.buf	= data,
		},
	};

	ret = sun7i_i2c_do_xfer(i2c, msg, 1);
	if (ret < 0)
		pr_debug("%s i2c xfer error: %d\n", __func__, ret);

	free(data);
	return ret;
	
}

#endif /* CONFIG_HARD_I2C */
