#include "athrs17_phy.h"
#define MII_BUSY 0x00000001
#define MII_WRITE 0x00000002

extern int phy_reg_write(uint32_t phy_addr, uint8_t reg, uint16_t data);
extern uint16_t phy_reg_read(uint32_t phy_addr, uint8_t reg);
#define     PHY_CONTROL_REG            0x0000	/*Control Register */
#define     PHY_STATUS_REG             0x0001	/*Status Register */
#define     PHY_ID_HI_REG              0x0002	/*PHY Identifier High Register */
#define     PHY_ID_LOW_REG             0x0003	/*PHY Identifier High Register */
#define     PHY_AN_ADV_REG             0x0004	/*Auto-Negotiation Advertisement Register */
#define     PHY_LNK_PART_ABl_REG       0x0005	/*Link Partner Ability Register (Base Page) */
#define     PHY_AN_EXP_REG             0x0006	/*Auto-Negotiation Expansion Register */
#define     PHY_AN_NXT_PAGE_TX_REG     0x0007	/*Next Page Transmit Register */
#define     PHY_LNK_PART_NXT_PAGE_REG  0x0008	/*Link Partner Next Page Register */
#define     PHY_1000BT_CTRL_REG        0x0009	/*1000BASE-T Control Register */
#define     PHY_1000BT_STATUS_REG      0x000a	/*1000BASE-T Status Register */
#define     PHY_SPECIFIC_CTRL_REG      0x0010	/*Phy specific control register */
#define     PHY_SPECIFIC_STATUS_REG    0x0011	/*Phy specific status register */
#define     PHY_INTERRUPT_ENABLE_REG   0x0012	/*Phy interrupt enable register */
#define     PHY_INTERRUPT_STATUS_REG   0x0013	/*Phy interrupt status register */
#define     PHY_EXT_PHY_SPC_CTRL       0x0014	/*Extended Phy specific control */
#define     PHY_RX_ERR_COUNTER         0x0015	/*Receive Error Counter */
#define     PHY_EXT_ADDR_CBL_DIAG      0x0016	/*Extended address for cable diagnostic register */
#define     PHY_LED_CONTROL            0x0018	/*LED Control */
#define     PHY_MAN_LED_OVERIDE        0x0019	/*Manual LED override register */
#define     PHY_EXT_PHY_SPC_CTRL2      0x001a	/*Extended Phy specific control 2 */
#define     PHY_EXT_PHY_SPC_STATUS     0x001b	/*Extended Phy specific status */
#define     PHY_CBL_DIAG_REG           0x001c	/*Cable diagnostic registers */

// Bits on PHY Register 1 (LOCAL Side Status) - RO:
#define PHY_STATUS_CAPPRESUPP   (1<< 6) // RO: '1'- Capable of MDIO Preamble Suppression
#define PHY_STATUS_AUTONEGDONE  (1<< 5) // RO: '1'- Auto-Negitiation Process Completed
#define PHY_STATUS_RMTFAULT     (1<< 4) // RO: '1'- Remote Fault. Latched High untill read
#define PHY_STATUS_CAPAUTONEG   (1<< 3) // RO: '1'- Capable of Auto-Negotiation Operation
#define PHY_STATUS_LINK_STATUS  (1<< 2) // RO: '1'- Link OK; '0'-Link Down. Latched Low untill read

#define PHY_Reg0                   0
#define PHY_Reg0_RESET        (1<<15) // WR: '1'-> SW Reset Phy. RD: Always '0'
#define PHY_Reg0_LOOPBACK     (1<<14) // RW: '0'- Disable; '1'- Enable Loopback
#define PHY_Reg0_SPEED        (1<<13) // RW: '0'- 10 Mbps; '1'- 100 Mbps
#define PHY_Reg0_AUTONEG      (1<<12) // RW: '0'- Disable; '1'- Enable Auto-Negotiation Process
#define PHY_Reg0_PWRDOWN      (1<<11) // RW: '0'- Disable; '1'- Enable Power Down mode
#define PHY_Reg0_ISOLATE      (1<<10) // RW: '0'- Disable; '1'- Enable PHY-MAC electrical Isolation
#define PHY_Reg0_AUTONEGRST   (1<< 9) // WR: '1'-> AutoNeg Restart. RD: Always '0'
#define PHY_Reg0_DUPLEX       (1<< 8) // RW: '0'- Half; '1'- Duplex Mode
#define PHY_Reg0_COLTEST      (1<< 7) // RW: '0'- Disable; '1'- Enable Collision Test
// bits 6 - 0 are Reserved, Always Read '0'.

// Bits on PHY Register 1 (LOCAL Side Status) - RO:
#define PHY_Reg1                   1
#define PHY_Reg1_CAPPRESUPP   (1<< 6) // RO: '1'- Capable of MDIO Preamble Suppression
#define PHY_Reg1_AUTONEGDONE  (1<< 5) // RO: '1'- Auto-Negitiation Process Completed
#define PHY_Reg1_RMTFAULT     (1<< 4) // RO: '1'- Remote Fault. Latched High untill read
#define PHY_Reg1_CAPAUTONEG   (1<< 3) // RO: '1'- Capable of Auto-Negotiation Operation
#define PHY_Reg1_LINK_STATUS  (1<< 2) // RO: '1'- Link OK; '0'-Link Down. Latched Low untill read

#define PHY_Reg2                   2 //PHY Identifier 0
#define PHY_Reg3                   3 //PHY Identifier 1
// Bits on PHY Registers 4 (Local  Side Auto-Negotiation Advertisment) - RW,
//                   and 5 (Remote Side Auto-Negotiation Advertisment) - RO.
// These Registers Are Send (4) and Received (5) during the Auto Negotiation Process,
// and Reflects each side Capabilities.
#define PHY_Reg4                    4
#define PHY_Reg5                    5
#define PHY_Reg4_5_PAUSE_OP    (1<<10)
#define PHY_Reg4_5_FD_100M     (1<< 8)
#define PHY_Reg4_5_HD_100M     (1<< 7)
#define PHY_Reg4_5_FD_10M      (1<< 6)
#define PHY_Reg4_5_HD_10M      (1<< 5)

#define PHY_Reg6                    6
#define PHY_Reg6_CapAutoNeg     (1<<0)

enum Mii_GEN_CTRL
{								/*    Description                bits        R/W  default value  */
	Mii_reset = 0x8000,
	Mii_Speed_10 = 0x0000,		/* 10   Mbps                    6:13           RW                      */
	Mii_Speed_100 = 0x2000,		/* 100  Mbps                    6:13           RW                      */
	Mii_Speed_1000 = 0x0040,	/* 1000 Mbit/s                  6:13           RW                      */

	Mii_Duplex = 0x0100,		/* Full Duplex mode             8                  RW                      */
	Mii_Manual_Master_Config = 0x0800,	/* Manual Master Config         11                 RW          */
	Mii_Loopback = 0x4000,		/* Enable Loop back             14                 RW                      */
	Mii_NoLoopback = 0x0000,	/* Enable Loop back             14                 RW                      */
};

enum Mii_Phy_Status
{
	Mii_phy_status_speed_10 = 0x0000,
	Mii_phy_status_speed_100 = 0x4000,
	Mii_phy_status_speed_1000 = 0x8000,

	Mii_phy_status_full_duplex = 0x2000,
	Mii_phy_status_half_duplex = 0x0000,
};

enum Mii_Link_Status
{
	LINKDOWN	= 0,
	LINKUP		= 1,
};

enum Mii_Duplex_Mode
{
	HALFDUPLEX = 0,
	FULLDUPLEX = 1,
};
enum Mii_Link_Speed
{
	SPEED10     = 10,
	SPEED100    = 100,
	SPEED1000   = 1000,
};

