/*
 * drivers/watchdog/sunxi_wdt.c
 * (C) Copyright 2010-2015
 * Reuuimlla Technology Co., Ltd. <www.reuuimllatech.com>
 * abai <abai@reuuimllatech.com>
 *
 * sun7i watchdog driver
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 */
#include <common.h>
#include <asm/io.h>

#include <watchdog.h>

#define MAX_TIMEOUT 	16 /* max 16 seconds */
#define IO_ADDRESS(x)		((x) + 0xf0000000)

#define SW_PA_TIMERC_IO_BASE              0x01c20c00
#define SW_VA_TIMERC_IO_BASE              (SW_PA_TIMERC_IO_BASE)

/* timer reg offset */
#define TMR_IRQ_EN_REG_OFF               0x0000
#define TMR_IRQ_STA_REG_OFF              0x0004

#define TMR_IRQ_EN_REG                   (SW_VA_TIMERC_IO_BASE + TMR_IRQ_EN_REG_OFF)
#define TMR_IRQ_STA_REG                  (SW_VA_TIMERC_IO_BASE + TMR_IRQ_STA_REG_OFF)

#define WDOG_CTRL_REG_OFF                0x0090
#define WDOG_MODE_REG_OFF                0x0094
#define WDOG_CTRL_REG                    (SW_VA_TIMERC_IO_BASE + WDOG_CTRL_REG_OFF)
#define WDOG_MODE_REG                   (SW_VA_TIMERC_IO_BASE + WDOG_MODE_REG_OFF)

static u8 wdt_is_active = 0;

struct sunxi_watchdog_interv{
	u32 	timeout;
	u32 	interv;
};

static const struct sunxi_watchdog_interv g_timeout_interv[] = {
	{.timeout = 0.5, .interv = 0b0000},
	{.timeout = 1  , .interv = 0b0001},
	{.timeout = 2  , .interv = 0b0010},
	{.timeout = 3  , .interv = 0b0011},
	{.timeout = 4  , .interv = 0b0100},
	{.timeout = 5  , .interv = 0b0101},
	{.timeout = 6  , .interv = 0b0110},
	{.timeout = 8  , .interv = 0b0111},
	{.timeout = 10 , .interv = 0b1000},
	{.timeout = 12 , .interv = 0b1001},
	{.timeout = 14 , .interv = 0b1010},
	{.timeout = 16 , .interv = 0b1011},
};

/* watchdog timeout in second */
static unsigned int g_timeout = MAX_TIMEOUT;

static void wdt_irq_en(u8 enable)
{
	u32 val = readl(TMR_IRQ_EN_REG);

	printf("%s, enable %d\n", __func__, enable);
	/* bit8: watcdog irq enable */
	val &= ~(1<<8);
	val |= (enable ? (1<<8) : (0<<8));
	writel(val, TMR_IRQ_EN_REG);
}

static void wdt_clr_irq_pend(void)
{
	u32 val = readl(TMR_IRQ_STA_REG);

	printf("%s\n", __func__);
	/* bit8: watcdog irq pending */
	if(val & (1<<8))
		writel(1<<8, TMR_IRQ_STA_REG);
}

static int timeout_to_interv(int timeout_in_sec)
{
	int 	temp;
	int 	array_sz;

	array_sz = ARRAY_SIZE(g_timeout_interv);
	if(timeout_in_sec > g_timeout_interv[array_sz - 1].timeout)
		return g_timeout_interv[array_sz - 1].interv;
	else if(timeout_in_sec < g_timeout_interv[0].timeout)
		return g_timeout_interv[0].interv;
	else {
		for(temp = 0; temp < array_sz; temp++) {
			if(timeout_in_sec > g_timeout_interv[temp].timeout)
				continue;
			else
				return g_timeout_interv[temp].interv;
		}
		printf("%s, line %d\n", __func__, __LINE__);
		return g_timeout_interv[array_sz - 1].interv; /* the largest one */
	}
}

static int interv_to_timeout(int interv)
{
	int 	temp;
	int 	array_sz;

	array_sz = ARRAY_SIZE(g_timeout_interv);
	if(interv > g_timeout_interv[array_sz - 1].interv)
		return g_timeout_interv[array_sz - 1].timeout;
	else if(array_sz < g_timeout_interv[0].interv)
		return g_timeout_interv[0].timeout;
	else {
		for(temp = 0; temp < array_sz; temp++) {
			if(interv > g_timeout_interv[temp].interv)
				continue;
			else
				return g_timeout_interv[temp].timeout;
		}
		printf("%s, line %d\n", __func__, __LINE__);
		return g_timeout_interv[array_sz - 1].timeout; /* the largest one */
	}
}

static void wdt_restart(void)
{
	printf("%s\n", __func__);
	/* bit0: watcdog restart */
	writel(((0xA57)<<1)|1 ,WDOG_CTRL_REG);
}

/**
 * wdt_set_tmout - set watchdog time out value
 * @timeout_in_sec:	time out to set in second
 *
 * Returns <0 if failed, otherwise return the actually timeout set in seconds.
 */
static int wdt_set_tmout(int timeout_in_sec)
{
	int 	temp = 0, temp2 = 0;
	int 	interv = 0;

	interv = timeout_to_interv(timeout_in_sec);
	printf("%s, timeout_in_sec %d, timeout_to_interv return interv 0x%08x\n",
		__func__, timeout_in_sec, interv);
	if(interv < 0)
		return interv;
	temp = readl(WDOG_MODE_REG);
	temp &= ~(0b1111 << 3);
	temp |= (interv << 3);
	writel(temp, WDOG_MODE_REG);

	temp2 = interv_to_timeout(interv);
	printf("%s, write 0x%08x to modreg 0x%08x, interv_to_timeout return timeout %d\n",
		__func__, temp, (u32)WDOG_MODE_REG, temp2);
	return temp2;
}

static void wdt_enable(u8 enable)
{
	int 	temp = readl(WDOG_MODE_REG);

	printf("%s %d\n", __func__, enable);
	/* bit0: watcdog enable */
	temp &= ~(1<<0);
	if(enable)
		temp |= 1<<0;
	writel(temp, WDOG_MODE_REG);
}

static void wdt_reset_enable(u8 enable)
{
	int 	temp = readl(WDOG_MODE_REG);

	printf("%s %d\n", __func__, enable);
	/* bit1: watcdog reset enable */
	temp &= ~(1<<1);
	if(enable)
		temp |= 1<<1;
	writel(temp, WDOG_MODE_REG);
}

void sunxi_watchdog_start(void)
{	
	if(wdt_is_active)
		return;

	int temp = wdt_set_tmout(g_timeout);

	if(temp < 0)
		return;
	g_timeout = temp;
	wdt_reset_enable(1); /* enable reset default */
	wdt_enable(1);
	wdt_is_active = 1;
}

static void sunxi_watchdog_kick(void)
{
	wdt_restart();
}

/**
 * watchdog_set_timeout - set watchdog time out value
 * @timeout:	time out to set in second
 *
 * Returns <0 if failed, 0 if success
 */
static int sunxi_watchdog_set_timeout(int timeout)
{
	int temp = wdt_set_tmout(timeout);

	if(temp < 0)
		return temp;
	g_timeout = temp;
	return 0;
}

int sunxi_watchdog_init(void)
{
	int 	temp = -1;
	
	/* disable watchdog */
	wdt_enable(0);
	/* disable irq */
	wdt_irq_en(0);
	/* clear irq pending */
	wdt_clr_irq_pend();
	/* set max timeout */
	temp = wdt_set_tmout(g_timeout);
	if(temp < 0)
		return temp;
	g_timeout = temp;

	return 0;
}

static int sunxi_wdt_release()
{
	wdt_enable(0);
	wdt_is_active = 0;
	return 0;
}


